#include "pathedge.h"

#include <fstream.h>
#include <stdlib.h>
#include "pathnode.h"
#include "pathplanner.h"
#include "util.h"


PathEdge *PathEdge::makeEdgePtr(PathNode *start, PathNode *end, 
				double maxPenFrac, DArray<MiniEdge>& subEdges,
				PathPlanner *pathPlanner,
				const Workspace *workspace) {
  assert(start != end);

  if (makePath(start, end, maxPenFrac, pathPlanner, workspace, 10, 
	       subEdges)) {
    double maxFrac = -100;
    for (int i = 0; i < subEdges.length(); i++) {
      if (subEdges[i].penFrac > maxFrac)
	maxFrac = subEdges[i].penFrac;
    }
    PathEdge *newEdge = new PathEdge(start, end, maxFrac, pathPlanner, workspace);
    assert(newEdge);
    return newEdge;
  }
  return NULL;
}

bool PathEdge::makeEdgePtr(PathNode *start, PathNode *end, double penFrac,
			   PathPlanner *pathPlanner, 
			   const Workspace *workspace) {
  DArray<MiniEdge> subEdges;
  PathEdge *newEdge = makeEdgePtr(start, end, penFrac, subEdges, 
				  pathPlanner, workspace);
  if (newEdge) {
    start->addEdge(newEdge, end);
    end->addEdge(newEdge, start);
    pathPlanner->addEdge(newEdge);
    newEdge->subEdges = subEdges;
  }
  
  return (newEdge != NULL);
}

void PathEdge::moveToFree(double newPenFrac) {
  int numTries = 0;
  DArray<PathNode *> newNodes;
  DArray<PathEdge *> newEdges;
  bool posConn = false, conn = false;
  PathNode *gNode = NULL;
  Point2 guess;
  int maxNumTries = 25;

  if (makeEdgePtr(from, to, newPenFrac, pathPlanner, workspace)) 
    return;

  PathNode *mid = PathNode::getNodePtr(pmid2(.5, from->getPos(), to->getPos()), newPenFrac, workspace);
  PathNode *newFrom = from;
  PathNode *newTo = to;

  if (mid && 0) {
    if (makeEdgePtr(from, mid, newPenFrac, pathPlanner, workspace))
      newFrom = mid;
    else if (makeEdgePtr(to, mid, newPenFrac, pathPlanner, workspace))
      newTo = mid;
  }
    

  if (from->getDist() > to->getDist()) {
    PathNode *temp = newFrom;
    newFrom = newTo;
    newTo = temp;
  }

  newNodes.append(newFrom);
  newNodes.append(newTo);
  Point2 reg = psub2(newTo->getPos(), newFrom->getPos());
  Point2 nor;
  nor.x = reg.y;
  nor.y = -reg.x;
  //nor = pmul2(pabs2(reg)/pabs2(nor), nor);



  while ((!gNode || !posConn || !(conn = connected(newFrom, newTo, newEdges, newNodes, true))) &&
	 numTries < maxNumTries) {

    gNode = NULL;

    bool guessed = false;

    int nodeTries = 0;
    while ((!guessed || 
	   !(gNode = PathNode::getNodePtr(guess, newPenFrac, workspace))) && 
    	   nodeTries++ < 4) {
      guessed = true;
      double place;
      place = drand48();
      place *= place;
      guess = padd2(pmid2(place, newFrom->getPos(), 
    			  newTo->getPos()),
    		    pmul2((.5-drand48()), nor));
    }
    if (gNode) {
      int nAdded = 0;
      for (int i = 0; i < newNodes.length(); i++) {
	DArray<MiniEdge> subEdges;
	PathEdge *newEdge = makeEdgePtr(gNode, newNodes[i], newPenFrac, 
					subEdges, pathPlanner, workspace);
	if (newEdge) {
	  if ((i == 0) || (i == 1))
	    posConn = true;
	  gNode->addEdge(newEdge, newNodes[i]);
	  newNodes[i]->addEdge(newEdge, gNode);
	  newEdges.append(newEdge);
	  nAdded++;
	}
      }
      if (nAdded)
	newNodes.append(gNode);
      else
	delete gNode;
    }
    numTries++;
  }

  if (conn) {
    for (int i=0; i < newEdges.length(); i++) {
      pathPlanner->addEdge(newEdges[i]);
    }
    for (int i=2; i < newNodes.length(); i++)
      if (true) {
	pathPlanner->addNode(newNodes[i]);
      }
    //cout << "+";
    //    cout << newNodes.length();
    return;
  }
  cout << "-";
  //  cout << "-" << newNodes.length() << endl;

  if (newNodes.length() < 0) {
    int no;
    cout << "Press a key... ";
    cin >> no;
    PathNode::saveNodesVect(newNodes, "edgeNodes.vect");
    PathEdge::saveEdgesVect(newEdges, "edgeEdges.vect");
  }
  for (int i=2; i < newNodes.length(); i++)
    delete newNodes[i];
}  

void PathEdge::removeNode(const PathNode *node) {
  if (from == node)
    from = NULL;
  if  (to == node)
    to = NULL;
}

PathEdge::~PathEdge() {
  if (from)
    from->removeEdge(this, to);
  if (to)
    to->removeEdge(this, from);
  pathPlanner->removeEdge(this);
}

PathNode *PathEdge::getNode(int nodeNum) {
  assert (nodeNum == 0 || nodeNum == 1);

  if (nodeNum == 0)
    return from;
  return to;
}


// Private methods
PathEdge::PathEdge(PathNode *start, PathNode *end, double myPenFrac, 
		   PathPlanner *myPathPlanner, const Workspace *myWorkspace):
  penFrac(myPenFrac), pathPlanner(myPathPlanner), workspace(myWorkspace)
{
    from = start; to = end;
}

bool PathEdge::freePath(double maxPenFrac) {
  return (penFrac < maxPenFrac);
}

bool PathEdge::makePath(PathNode *start, PathNode *end, double maxPenFrac, 
			PathPlanner *myPlanner, const Workspace *workspace, 
			int depthLeft, DArray<MiniEdge>& subEdges) {

  double farDist = max(start->getDist(), end->getDist());
  double sepDist = start->distToNode(end);
  double penFrac = ((farDist-sepDist)/(start->getDist() < end->getDist() ? 
				       end->getMaxPen() : start->getMaxPen()));

  if (penFrac < maxPenFrac) {
    MiniEdge newEdge;
    //newEdge.from = start;
    //newEdge.to = end;
    newEdge.penFrac = penFrac;
    subEdges.append(newEdge);
    return true;
  } else if (depthLeft > 0) {
    PathNode *midNode;
    midNode = PathNode::getNodePtr(pmid2(.5, start->getPos(), 
					 end->getPos()), maxPenFrac, 
				   workspace);
    if (midNode) {
      bool pa = makePath(start, midNode, maxPenFrac, myPlanner, workspace, 
			 depthLeft-1, subEdges);
      bool pb = makePath(midNode, end, maxPenFrac, myPlanner, workspace, 
			 depthLeft-1, subEdges);
      delete midNode;
      return (pa && pb);
    } else {
      return false;
    }
  } else {
    return false;
  }
}

int PathEdge::numSegments() const {
  return 1;
}

bool PathEdge::connected(PathNode* from, PathNode* to, DArray<PathEdge *>& edges, DArray<PathNode *>& nodes, bool destructive) {
  DArray<PathNode *> visited;
  bool conn;
  
  visited.append(from);
  conn = connectedHelp(from, to, edges, visited);
  if (destructive && conn) {
    int i = 0;
    while (i < edges.length()) {
      if (!visited.contains(edges[i]->from) || 
	  !visited.contains(edges[i]->to))
	edges.removeAt(i);
      else
	i++;
    }
    i = 0;
    while (i < nodes.length()) {
      if (!visited.contains(nodes[i])) {
	delete nodes[i];
	nodes.removeAt(i);
      } else {
	i++;
      }
    }
  }

  if (!destructive)
    PathNode::saveNodesVect(visited, "nodes_v.vect");

  return conn;
}

bool PathEdge::connectedHelp(PathNode* start, PathNode* end, 
			  DArray<PathEdge *>& edges, 
			  DArray<PathNode *>& visited) {
  if (start == end)
    return true;

  for (int i = 0; i < edges.length(); i++) {
    PathNode *interm = NULL;
      if (edges[i]->from == start) {
	interm = edges[i]->to;
      }
      else if (edges[i]->to == start) {
	interm = edges[i]->from;
      }
      if (interm && !visited.contains(interm)) {
	visited.append(interm);
	//cout << visited.length();
	if (connectedHelp(interm, end, edges, visited)) {
	  return true;
	} else {
	  //visited.remove(interm);
	}
    }
  }   
   

  return false;
}
      

ostream& operator<<(ostream& out, const PathEdge& pathEdge) {
  Point2 pos1 = pathEdge.from->getPos();
  Point2 pos2 = pathEdge.to->getPos();
  out << pos1.x << " " << pos1.y << " 0\n";
  out << pos2.x << " " << pos2.y << " 0\n";

  return out;
}

void PathEdge::saveEdgesVect(DArray<PathEdge *>& edges, const char *outFileName) { 
  ofstream outfile(outFileName);
  assert(outfile);

  outfile << "VECT\n";
  outfile << edges.length() << " " << edges.length()*2 << " 0\n";
  for (int i = 0; i < edges.length(); i++)
    outfile << "2 ";
  outfile << "\n";
  for (int i = 0; i < edges.length(); i++)
    outfile << "0 ";
  outfile << "\n";
  for (int i = 0; i < edges.length(); i++) {
    Point2 pos1 = edges[i]->getNode(0)->getPos();
    Point2 pos2 = edges[i]->getNode(1)->getPos();
    outfile << pos1.x << " " << pos1.y << " 0\n";
    outfile << pos2.x << " " << pos2.y << " 0\n";
    }

}
